@@ -1,26 +1,52 @@
+from __future__ import annotations
+
 import datetime
 import warnings
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Generic,
+    Hashable,
+    Iterator,
+    Literal,
+    Mapping,
+    Sequence,
+    TypeVar,
+    Union,
+    cast,
+)
 
 import numpy as np
 import pandas as pd
 
 from . import dtypes, duck_array_ops, nputils, ops
 from ._reductions import DataArrayGroupByReductions, DatasetGroupByReductions
+from .alignment import align
 from .arithmetic import DataArrayGroupbyArithmetic, DatasetGroupbyArithmetic
+from .common import ImplementsArrayReduce, ImplementsDatasetReduce
 from .concat import concat
 from .formatting import format_array_flat
-from .indexes import propagate_indexes
-from .options import _get_keep_attrs
-from .pycompat import integer_types
-from .utils import (
-    either_dict_or_kwargs,
-    hashable,
-    is_scalar,
-    maybe_wrap_array,
-    peek_at,
+from .indexes import (
+    create_default_index_implicit,
+    filter_indexes_from_coords,
     safe_cast_to_index,
 )
-from .variable import IndexVariable, Variable, as_variable
+from .ops import IncludeCumMethods
+from .options import _get_keep_attrs
+from .pycompat import integer_types
+from .types import Dims, QuantileMethods, T_Xarray
+from .utils import either_dict_or_kwargs, hashable, is_scalar, maybe_wrap_array, peek_at
+from .variable import IndexVariable, Variable
+
+if TYPE_CHECKING:
+    from numpy.typing import ArrayLike
+
+    from .dataarray import DataArray
+    from .dataset import Dataset
+    from .utils import Frozen
+
+    GroupKey = Any
 
 
 def check_reduce_dims(reduce_dims, dimensions):
@@ -35,14 +61,16 @@ def check_reduce_dims(reduce_dims, dimensions):
             )
 
 
-def unique_value_groups(ar, sort=True):
+def unique_value_groups(
+    ar, sort: bool = True
+) -> tuple[np.ndarray | pd.Index, list[list[int]]]:
     """Group an array by its unique values.
 
     Parameters
     ----------
     ar : array-like
         Input array. This will be flattened if it is not already 1-D.
-    sort : bool, optional
+    sort : bool, default: True
         Whether or not to sort unique values.
 
     Returns
@@ -54,7 +82,9 @@ def unique_value_groups(ar, sort=True):
         the corresponding value in `unique_values`.
     """
     inverse, values = pd.factorize(ar, sort=sort)
-    groups = [[] for _ in range(len(values))]
+    if isinstance(values, pd.MultiIndex):
+        values.names = ar.names
+    groups: list[list[int]] = [[] for _ in range(len(values))]
     for n, g in enumerate(inverse):
         if g >= 0:
             # pandas uses -1 to mark NaN, but doesn't include them in values
@@ -153,25 +183,29 @@ class _DummyGroup:
 
     __slots__ = ("name", "coords", "size")
 
-    def __init__(self, obj, name, coords):
+    def __init__(self, obj: T_Xarray, name: Hashable, coords) -> None:
         self.name = name
         self.coords = coords
         self.size = obj.sizes[name]
 
     @property
-    def dims(self):
+    def dims(self) -> tuple[Hashable]:
         return (self.name,)
 
     @property
-    def ndim(self):
+    def ndim(self) -> Literal[1]:
         return 1
 
     @property
-    def values(self):
+    def values(self) -> range:
         return range(self.size)
 
     @property
-    def shape(self):
+    def data(self) -> range:
+        return range(self.size)
+
+    @property
+    def shape(self) -> tuple[int]:
         return (self.size,)
 
     def __getitem__(self, key):
@@ -180,24 +214,36 @@ class _DummyGroup:
         return self.values[key]
 
 
-def _ensure_1d(group, obj):
-    if group.ndim != 1:
+T_Group = TypeVar("T_Group", bound=Union["DataArray", "IndexVariable", _DummyGroup])
+
+
+def _ensure_1d(
+    group: T_Group, obj: T_Xarray
+) -> tuple[T_Group, T_Xarray, Hashable | None, list[Hashable]]:
+    # 1D cases: do nothing
+    from .dataarray import DataArray
+
+    if isinstance(group, (IndexVariable, _DummyGroup)) or group.ndim == 1:
+        return group, obj, None, []
+
+    if isinstance(group, DataArray):
         # try to stack the dims of the group into a single dim
         orig_dims = group.dims
-        stacked_dim = "stacked_" + "_".join(orig_dims)
+        stacked_dim = "stacked_" + "_".join(map(str, orig_dims))
         # these dimensions get created by the stack operation
         inserted_dims = [dim for dim in group.dims if dim not in group.coords]
         # the copy is necessary here, otherwise read only array raises error
         # in pandas: https://github.com/pydata/pandas/issues/12813
-        group = group.stack(**{stacked_dim: orig_dims}).copy()
-        obj = obj.stack(**{stacked_dim: orig_dims})
-    else:
-        stacked_dim = None
-        inserted_dims = []
-    return group, obj, stacked_dim, inserted_dims
+        newgroup = group.stack({stacked_dim: orig_dims}).copy()
+        newobj = obj.stack({stacked_dim: orig_dims})
+        return cast(T_Group, newgroup), newobj, stacked_dim, inserted_dims
+
+    raise TypeError(
+        f"group must be DataArray, IndexVariable or _DummyGroup, got {type(group)!r}."
+    )
 
 
-def _unique_and_monotonic(group):
+def _unique_and_monotonic(group: T_Group) -> bool:
     if isinstance(group, _DummyGroup):
         return True
     index = safe_cast_to_index(group)
@@ -230,7 +276,7 @@ def _apply_loffset(grouper, result):
     grouper.loffset = None
 
 
-class GroupBy:
+class GroupBy(Generic[T_Xarray]):
     """A object that implements the split-apply-combine pattern.
 
     Modeled after `pandas.GroupBy`. The `GroupBy` object can be iterated over
@@ -259,27 +305,34 @@ class GroupBy:
         "_stacked_dim",
         "_unique_coord",
         "_dims",
+        "_sizes",
+        "_squeeze",
+        # Save unstacked object for flox
+        "_original_obj",
+        "_original_group",
+        "_bins",
     )
+    _obj: T_Xarray
 
     def __init__(
         self,
-        obj,
-        group,
-        squeeze=False,
-        grouper=None,
-        bins=None,
-        restore_coord_dims=True,
-        cut_kwargs=None,
-    ):
+        obj: T_Xarray,
+        group: Hashable | DataArray | IndexVariable,
+        squeeze: bool = False,
+        grouper: pd.Grouper | None = None,
+        bins: ArrayLike | None = None,
+        restore_coord_dims: bool = True,
+        cut_kwargs: Mapping[Any, Any] | None = None,
+    ) -> None:
         """Create a GroupBy object
 
         Parameters
         ----------
         obj : Dataset or DataArray
             Object to group.
-        group : DataArray
-            Array with the group values.
-        squeeze : bool, optional
+        group : Hashable, DataArray or Index
+            Array with the group values or name of the variable.
+        squeeze : bool, default: False
             If "group" is a coordinate of object, `squeeze` controls whether
             the subarrays have a dimension of length 1 along that coordinate or
             if the dimension is squeezed out.
@@ -291,7 +344,7 @@ class GroupBy:
         restore_coord_dims : bool, default: True
             If True, also restore the dimension order of multi-dimensional
             coordinates.
-        cut_kwargs : dict, optional
+        cut_kwargs : dict-like, optional
             Extra keyword arguments to pass to `pandas.cut`
 
         """
@@ -306,7 +359,7 @@ class GroupBy:
             if not hashable(group):
                 raise TypeError(
                     "`group` must be an xarray.DataArray or the "
-                    "name of an xarray variable or dimension."
+                    "name of an xarray variable or dimension. "
                     f"Received {group!r} instead."
                 )
             group = obj[group]
@@ -320,6 +373,10 @@ class GroupBy:
         if getattr(group, "name", None) is None:
             group.name = "group"
 
+        self._original_obj: T_Xarray = obj
+        self._original_group = group
+        self._bins = bins
+
         group, obj, stacked_dim, inserted_dims = _ensure_1d(group, obj)
         (group_dim,) = group.dims
 
@@ -336,11 +393,13 @@ class GroupBy:
         if bins is not None:
             if duck_array_ops.isnull(bins).all():
                 raise ValueError("All bin edges are NaN.")
-            binned = pd.cut(group.values, bins, **cut_kwargs)
-            new_dim_name = group.name + "_bins"
-            group = DataArray(binned, group.coords, name=new_dim_name)
+            binned, bins = pd.cut(group.values, bins, **cut_kwargs, retbins=True)
+            new_dim_name = str(group.name) + "_bins"
+            group = DataArray(binned, getattr(group, "coords", None), name=new_dim_name)
             full_index = binned.categories
 
+        group_indices: list[slice] | list[list[int]] | np.ndarray
+        unique_coord: DataArray | IndexVariable | _DummyGroup
         if grouper is not None:
             index = safe_cast_to_index(group)
             if not index.is_monotonic_increasing:
@@ -354,14 +413,15 @@ class GroupBy:
             unique_coord = IndexVariable(group.name, first_items.index)
         elif group.dims == (group.name,) and _unique_and_monotonic(group):
             # no need to factorize
-            group_indices = np.arange(group.size)
             if not squeeze:
                 # use slices to do views instead of fancy indexing
                 # equivalent to: group_indices = group_indices.reshape(-1, 1)
-                group_indices = [slice(i, i + 1) for i in group_indices]
+                group_indices = [slice(i, i + 1) for i in range(group.size)]
+            else:
+                group_indices = np.arange(group.size)
             unique_coord = group
         else:
-            if group.isnull().any():
+            if isinstance(group, DataArray) and group.isnull().any():
                 # drop any NaN valued groups.
                 # also drop obj values where group was NaN
                 # Use where instead of reindex to account for duplicate coordinate labels.
@@ -387,7 +447,7 @@ class GroupBy:
                 )
 
         # specification for the groupby operation
-        self._obj = obj
+        self._obj: T_Xarray = obj
         self._group = group
         self._group_dim = group_dim
         self._group_indices = group_indices
@@ -396,22 +456,56 @@ class GroupBy:
         self._inserted_dims = inserted_dims
         self._full_index = full_index
         self._restore_coord_dims = restore_coord_dims
+        self._bins = bins
+        self._squeeze = squeeze
 
         # cached attributes
-        self._groups = None
-        self._dims = None
+        self._groups: dict[GroupKey, slice | int | list[int]] | None = None
+        self._dims: tuple[Hashable, ...] | Frozen[Hashable, int] | None = None
+        self._sizes: Frozen[Hashable, int] | None = None
 
     @property
-    def dims(self):
-        if self._dims is None:
-            self._dims = self._obj.isel(
-                **{self._group_dim: self._group_indices[0]}
-            ).dims
+    def sizes(self) -> Frozen[Hashable, int]:
+        """Ordered mapping from dimension names to lengths.
 
-        return self._dims
+        Immutable.
+
+        See Also
+        --------
+        DataArray.sizes
+        Dataset.sizes
+        """
+        if self._sizes is None:
+            self._sizes = self._obj.isel(
+                {self._group_dim: self._group_indices[0]}
+            ).sizes
+
+        return self._sizes
+
+    def map(
+        self,
+        func: Callable,
+        args: tuple[Any, ...] = (),
+        shortcut: bool | None = None,
+        **kwargs: Any,
+    ) -> T_Xarray:
+        raise NotImplementedError()
+
+    def reduce(
+        self,
+        func: Callable[..., Any],
+        dim: Dims | ellipsis = None,
+        *,
+        axis: int | Sequence[int] | None = None,
+        keep_attrs: bool | None = None,
+        keepdims: bool = False,
+        shortcut: bool = True,
+        **kwargs: Any,
+    ) -> T_Xarray:
+        raise NotImplementedError()
 
     @property
-    def groups(self):
+    def groups(self) -> dict[GroupKey, slice | int | list[int]]:
         """
         Mapping from group labels to indices. The indices can be used to index the underlying object.
         """
@@ -420,19 +514,19 @@ class GroupBy:
             self._groups = dict(zip(self._unique_coord.values, self._group_indices))
         return self._groups
 
-    def __getitem__(self, key):
+    def __getitem__(self, key: GroupKey) -> T_Xarray:
         """
         Get DataArray or Dataset corresponding to a particular group label.
         """
         return self._obj.isel({self._group_dim: self.groups[key]})
 
-    def __len__(self):
+    def __len__(self) -> int:
         return self._unique_coord.size
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[tuple[GroupKey, T_Xarray]]:
         return zip(self._unique_coord.values, self._iter_grouped())
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "{}, grouped over {!r}\n{!r} groups with labels {}.".format(
             self.__class__.__name__,
             self._unique_coord.name,
@@ -454,10 +548,10 @@ class GroupBy:
             first_items = first_items.dropna()
         return full_index, first_items
 
-    def _iter_grouped(self):
+    def _iter_grouped(self) -> Iterator[T_Xarray]:
         """Iterate over each element in this group"""
         for indices in self._group_indices:
-            yield self._obj.isel(**{self._group_dim: indices})
+            yield self._obj.isel({self._group_dim: indices})
 
     def _infer_concat_args(self, applied_example):
         if self._group_dim in applied_example.dims:
@@ -469,38 +563,73 @@ class GroupBy:
         (dim,) = coord.dims
         if isinstance(coord, _DummyGroup):
             coord = None
+        coord = getattr(coord, "variable", coord)
         return coord, dim, positions
 
     def _binary_op(self, other, f, reflexive=False):
+        from .dataarray import DataArray
+        from .dataset import Dataset
+
         g = f if not reflexive else lambda x, y: f(y, x)
-        applied = self._yield_binary_applied(g, other)
-        return self._combine(applied)
 
-    def _yield_binary_applied(self, func, other):
-        dummy = None
+        if self._bins is None:
+            obj = self._original_obj
+            group = self._original_group
+            dims = group.dims
+        else:
+            obj = self._maybe_unstack(self._obj)
+            group = self._maybe_unstack(self._group)
+            dims = (self._group_dim,)
 
-        for group_value, obj in self:
-            try:
-                other_sel = other.sel(**{self._group.name: group_value})
-            except AttributeError:
-                raise TypeError(
-                    "GroupBy objects only support binary ops "
-                    "when the other argument is a Dataset or "
-                    "DataArray"
+        if isinstance(group, _DummyGroup):
+            group = obj[group.name]
+            coord = group
+        else:
+            coord = self._unique_coord
+            if not isinstance(coord, DataArray):
+                coord = DataArray(self._unique_coord)
+        name = group.name
+
+        if not isinstance(other, (Dataset, DataArray)):
+            raise TypeError(
+                "GroupBy objects only support binary ops "
+                "when the other argument is a Dataset or "
+                "DataArray"
+            )
+
+        if name not in other.dims:
+            raise ValueError(
+                "incompatible dimensions for a grouped "
+                f"binary operation: the group variable {name!r} "
+                "is not a dimension on the other argument"
+            )
+
+        # Broadcast out scalars for backwards compatibility
+        # TODO: get rid of this when fixing GH2145
+        for var in other.coords:
+            if other[var].ndim == 0:
+                other[var] = (
+                    other[var].drop_vars(var).expand_dims({name: other.sizes[name]})
                 )
-            except (KeyError, ValueError):
-                if self._group.name not in other.dims:
-                    raise ValueError(
-                        "incompatible dimensions for a grouped "
-                        f"binary operation: the group variable {self._group.name!r} "
-                        "is not a dimension on the other argument"
-                    )
-                if dummy is None:
-                    dummy = _dummy_copy(other)
-                other_sel = dummy
 
-            result = func(obj, other_sel)
-            yield result
+        other, _ = align(other, coord, join="outer")
+        expanded = other.sel({name: group})
+
+        result = g(obj, expanded)
+
+        if group.ndim > 1:
+            # backcompat:
+            # TODO: get rid of this when fixing GH2145
+            for var in set(obj.coords) - set(obj.xindexes):
+                if set(obj[var].dims) < set(group.dims):
+                    result[var] = obj[var].reset_coords(drop=True).broadcast_like(group)
+
+        if isinstance(result, Dataset) and isinstance(obj, Dataset):
+            for var in set(result):
+                for d in dims:
+                    if d not in obj[var].dims:
+                        result[var] = result[var].transpose(d, ...)
+        return result
 
     def _maybe_restore_empty_groups(self, combined):
         """Our index contained empty groups (e.g., from a resampling). If we
@@ -519,10 +648,142 @@ class GroupBy:
             for dim in self._inserted_dims:
                 if dim in obj.coords:
                     del obj.coords[dim]
-            obj._indexes = propagate_indexes(obj._indexes, exclude=self._inserted_dims)
+            obj._indexes = filter_indexes_from_coords(obj._indexes, set(obj.coords))
         return obj
 
-    def fillna(self, value):
+    def _flox_reduce(
+        self,
+        dim: Dims | ellipsis,
+        keep_attrs: bool | None = None,
+        **kwargs: Any,
+    ):
+        """Adaptor function that translates our groupby API to that of flox."""
+        from flox.xarray import xarray_reduce
+
+        from .dataset import Dataset
+
+        obj = self._original_obj
+
+        if keep_attrs is None:
+            keep_attrs = _get_keep_attrs(default=True)
+
+        # preserve current strategy (approximately) for dask groupby.
+        # We want to control the default anyway to prevent surprises
+        # if flox decides to change its default
+        kwargs.setdefault("method", "split-reduce")
+
+        numeric_only = kwargs.pop("numeric_only", None)
+        if numeric_only:
+            non_numeric = {
+                name: var
+                for name, var in obj.data_vars.items()
+                if not (np.issubdtype(var.dtype, np.number) or (var.dtype == np.bool_))
+            }
+        else:
+            non_numeric = {}
+
+        # weird backcompat
+        # reducing along a unique indexed dimension with squeeze=True
+        # should raise an error
+        if (
+            dim is None or dim == self._group.name
+        ) and self._group.name in obj.xindexes:
+            index = obj.indexes[self._group.name]
+            if index.is_unique and self._squeeze:
+                raise ValueError(f"cannot reduce over dimensions {self._group.name!r}")
+
+        # group is only passed by resample
+        group = kwargs.pop("group", None)
+        if group is None:
+            if isinstance(self._original_group, _DummyGroup):
+                group = self._original_group.name
+            else:
+                group = self._original_group
+
+        unindexed_dims: tuple[str, ...] = tuple()
+        if isinstance(group, str):
+            if group in obj.dims and group not in obj._indexes and self._bins is None:
+                unindexed_dims = (group,)
+            group = self._original_obj[group]
+
+        parsed_dim: tuple[Hashable, ...]
+        if isinstance(dim, str):
+            parsed_dim = (dim,)
+        elif dim is None:
+            parsed_dim = group.dims
+        elif dim is ...:
+            parsed_dim = tuple(self._original_obj.dims)
+        else:
+            parsed_dim = tuple(dim)  # type:ignore[arg-type]
+
+        # Do this so we raise the same error message whether flox is present or not.
+        # Better to control it here than in flox.
+        if any(
+            d not in group.dims and d not in self._original_obj.dims for d in parsed_dim
+        ):
+            raise ValueError(f"cannot reduce over dimensions {dim}.")
+
+        expected_groups: tuple[np.ndarray | Any, ...]
+        isbin: bool | Sequence[bool]
+        if self._bins is not None:
+            # TODO: fix this; When binning by time, self._bins is a DatetimeIndex
+            expected_groups = (np.array(self._bins),)
+            isbin = (True,)
+            # This is an annoying hack. Xarray returns np.nan
+            # when there are no observations in a bin, instead of 0.
+            # We can fake that here by forcing min_count=1.
+            if kwargs["func"] == "count":
+                if "fill_value" not in kwargs or kwargs["fill_value"] is None:
+                    kwargs["fill_value"] = np.nan
+                    # note min_count makes no sense in the xarray world
+                    # as a kwarg for count, so this should be OK
+                    kwargs["min_count"] = 1
+            # empty bins have np.nan regardless of dtype
+            # flox's default would not set np.nan for integer dtypes
+            kwargs.setdefault("fill_value", np.nan)
+        else:
+            expected_groups = (self._unique_coord.values,)
+            isbin = False
+
+        result = xarray_reduce(
+            self._original_obj.drop_vars(non_numeric),
+            group,
+            dim=parsed_dim,  # type:ignore[arg-type]  # https://github.com/xarray-contrib/flox/issues/96
+            expected_groups=expected_groups,
+            isbin=isbin,
+            keep_attrs=keep_attrs,
+            **kwargs,
+        )
+
+        # Ignore error when the groupby reduction is effectively
+        # a reduction of the underlying dataset
+        result = result.drop_vars(unindexed_dims, errors="ignore")
+
+        # broadcast and restore non-numeric data variables (backcompat)
+        for name, var in non_numeric.items():
+            if all(d not in var.dims for d in parsed_dim):
+                result[name] = var.variable.set_dims(
+                    (group.name,) + var.dims, (result.sizes[group.name],) + var.shape
+                )
+
+        if self._bins is not None:
+            # bins provided to flox are at full precision
+            # the bin edge labels have a default precision of 3
+            # reassign to fix that.
+            assert self._full_index is not None
+            new_coord = [
+                pd.Interval(inter.left, inter.right) for inter in self._full_index
+            ]
+            result[self._group.name] = new_coord
+            # Fix dimension order when binning a dimension coordinate
+            # Needed as long as we do a separate code path for pint;
+            # For some reason Datasets and DataArrays behave differently!
+            if isinstance(self._obj, Dataset) and self._group_dim in self._obj.dims:
+                result = result.transpose(self._group.name, ...)
+
+        return result
+
+    def fillna(self, value: Any) -> T_Xarray:
         """Fill missing values in this object by group.
 
         This operation follows the normal broadcasting and alignment rules that
@@ -550,13 +811,13 @@ class GroupBy:
 
     def quantile(
         self,
-        q,
-        dim=None,
-        method="linear",
-        keep_attrs=None,
-        skipna=True,
-        interpolation=None,
-    ):
+        q: ArrayLike,
+        dim: Dims = None,
+        method: QuantileMethods = "linear",
+        keep_attrs: bool | None = None,
+        skipna: bool | None = None,
+        interpolation: QuantileMethods | None = None,
+    ) -> T_Xarray:
         """Compute the qth quantile over each array in the groups and
         concatenate them together into a new array.
 
@@ -565,7 +826,7 @@ class GroupBy:
         q : float or sequence of float
             Quantile to compute, which must be between 0 and 1
             inclusive.
-        dim : ..., str or sequence of str, optional
+        dim : str or Iterable of Hashable, optional
             Dimension(s) over which to apply quantile.
             Defaults to the grouped dimension.
         method : str, default: "linear"
@@ -595,9 +856,15 @@ class GroupBy:
             an asterix require numpy version 1.22 or newer. The "method" argument was
             previously called "interpolation", renamed in accordance with numpy
             version 1.22.0.
-
-        skipna : bool, optional
-            Whether to skip missing values when aggregating.
+        keep_attrs : bool or None, default: None
+            If True, the dataarray's attributes (`attrs`) will be copied from
+            the original object to the new one.  If False, the new
+            object will be returned without attributes.
+        skipna : bool or None, default: None
+            If True, skip missing values (as marked by NaN). By default, only
+            skips missing values for float dtypes; other dtypes either do not
+            have a sentinel missing value (int) or skipna=True has not been
+            implemented (object, datetime64 or timedelta64).
 
         Returns
         -------
@@ -669,9 +936,9 @@ class GroupBy:
            The American Statistician, 50(4), pp. 361-365, 1996
         """
         if dim is None:
-            dim = self._group_dim
+            dim = (self._group_dim,)
 
-        out = self.map(
+        return self.map(
             self._obj.__class__.quantile,
             shortcut=False,
             q=q,
@@ -681,9 +948,8 @@ class GroupBy:
             skipna=skipna,
             interpolation=interpolation,
         )
-        return out
 
-    def where(self, cond, other=dtypes.NA):
+    def where(self, cond, other=dtypes.NA) -> T_Xarray:
         """Return elements from `self` or `other` depending on `cond`.
 
         Parameters
@@ -711,13 +977,15 @@ class GroupBy:
             return self._obj
         if keep_attrs is None:
             keep_attrs = _get_keep_attrs(default=True)
-        return self.reduce(op, self._group_dim, skipna=skipna, keep_attrs=keep_attrs)
+        return self.reduce(
+            op, dim=[self._group_dim], skipna=skipna, keep_attrs=keep_attrs
+        )
 
-    def first(self, skipna=None, keep_attrs=None):
+    def first(self, skipna: bool | None = None, keep_attrs: bool | None = None):
         """Return the first element of each group along the group dimension"""
         return self._first_or_last(duck_array_ops.first, skipna, keep_attrs)
 
-    def last(self, skipna=None, keep_attrs=None):
+    def last(self, skipna: bool | None = None, keep_attrs: bool | None = None):
         """Return the last element of each group along the group dimension"""
         return self._first_or_last(duck_array_ops.last, skipna, keep_attrs)
 
@@ -742,10 +1010,18 @@ def _maybe_reorder(xarray_obj, dim, positions):
         return xarray_obj[{dim: order}]
 
 
-class DataArrayGroupByBase(GroupBy, DataArrayGroupbyArithmetic):
+class DataArrayGroupByBase(GroupBy["DataArray"], DataArrayGroupbyArithmetic):
     """GroupBy object specialized to grouping DataArray objects"""
 
     __slots__ = ()
+    _dims: tuple[Hashable, ...] | None
+
+    @property
+    def dims(self) -> tuple[Hashable, ...]:
+        if self._dims is None:
+            self._dims = self._obj.isel({self._group_dim: self._group_indices[0]}).dims
+
+        return self._dims
 
     def _iter_grouped_shortcut(self):
         """Fast version of `_iter_grouped` that yields Variables without
@@ -760,11 +1036,13 @@ class DataArrayGroupByBase(GroupBy, DataArrayGroupbyArithmetic):
         # speed things up, but it's not very interpretable and there are much
         # faster alternatives (e.g., doing the grouped aggregation in a
         # compiled language)
+        # TODO: benbovy - explicit indexes: this fast implementation doesn't
+        # create an explicit index for the stacked dim coordinate
         stacked = Variable.concat(applied, dim, shortcut=True)
         reordered = _maybe_reorder(stacked, dim, positions)
         return self._obj._replace_maybe_drop_dims(reordered)
 
-    def _restore_dim_order(self, stacked):
+    def _restore_dim_order(self, stacked: DataArray) -> DataArray:
         def lookup_order(dimension):
             if dimension == self._group.name:
                 (dimension,) = self._group.dims
@@ -777,7 +1055,13 @@ class DataArrayGroupByBase(GroupBy, DataArrayGroupbyArithmetic):
         new_order = sorted(stacked.dims, key=lookup_order)
         return stacked.transpose(*new_order, transpose_coords=self._restore_coord_dims)
 
-    def map(self, func, shortcut=False, args=(), **kwargs):
+    def map(
+        self,
+        func: Callable[..., DataArray],
+        args: tuple[Any, ...] = (),
+        shortcut: bool | None = None,
+        **kwargs: Any,
+    ) -> DataArray:
         """Apply a function to each array in the group and concatenate them
         together into a new array.
 
@@ -816,7 +1100,7 @@ class DataArrayGroupByBase(GroupBy, DataArrayGroupbyArithmetic):
 
         Returns
         -------
-        applied : DataArray or DataArray
+        applied : DataArray
             The result of splitting, applying and combining this array.
         """
         grouped = self._iter_grouped_shortcut() if shortcut else self._iter_grouped()
@@ -851,20 +1135,26 @@ class DataArrayGroupByBase(GroupBy, DataArrayGroupbyArithmetic):
         if isinstance(combined, type(self._obj)):
             # only restore dimension order for arrays
             combined = self._restore_dim_order(combined)
-        # assign coord when the applied function does not return that coord
+        # assign coord and index when the applied function does not return that coord
         if coord is not None and dim not in applied_example.dims:
-            if shortcut:
-                coord_var = as_variable(coord)
-                combined._coords[coord.name] = coord_var
-            else:
-                combined.coords[coord.name] = coord
+            index, index_vars = create_default_index_implicit(coord)
+            indexes = {k: index for k in index_vars}
+            combined = combined._overwrite_indexes(indexes, index_vars)
         combined = self._maybe_restore_empty_groups(combined)
         combined = self._maybe_unstack(combined)
         return combined
 
     def reduce(
-        self, func, dim=None, axis=None, keep_attrs=None, shortcut=True, **kwargs
-    ):
+        self,
+        func: Callable[..., Any],
+        dim: Dims | ellipsis = None,
+        *,
+        axis: int | Sequence[int] | None = None,
+        keep_attrs: bool | None = None,
+        keepdims: bool = False,
+        shortcut: bool = True,
+        **kwargs: Any,
+    ) -> DataArray:
         """Reduce the items in this group by applying `func` along some
         dimension(s).
 
@@ -874,8 +1164,9 @@ class DataArrayGroupByBase(GroupBy, DataArrayGroupbyArithmetic):
             Function which can be called in the form
             `func(x, axis=axis, **kwargs)` to return the result of collapsing
             an np.ndarray over an integer valued axis.
-        dim : ..., str or sequence of str, optional
-            Dimension(s) over which to apply `func`.
+        dim : "...", str, Iterable of Hashable or None, optional
+            Dimension(s) over which to apply `func`. If None, apply over the
+            groupby dimension, if "..." apply over all dimensions.
         axis : int or sequence of int, optional
             Axis(es) over which to apply `func`. Only one of the 'dimension'
             and 'axis' arguments can be supplied. If neither are supplied, then
@@ -894,28 +1185,55 @@ class DataArrayGroupByBase(GroupBy, DataArrayGroupbyArithmetic):
             removed.
         """
         if dim is None:
-            dim = self._group_dim
+            dim = [self._group_dim]
 
         if keep_attrs is None:
-            keep_attrs = _get_keep_attrs(default=False)
+            keep_attrs = _get_keep_attrs(default=True)
 
-        def reduce_array(ar):
-            return ar.reduce(func, dim, axis, keep_attrs=keep_attrs, **kwargs)
+        def reduce_array(ar: DataArray) -> DataArray:
+            return ar.reduce(
+                func=func,
+                dim=dim,
+                axis=axis,
+                keep_attrs=keep_attrs,
+                keepdims=keepdims,
+                **kwargs,
+            )
 
         check_reduce_dims(dim, self.dims)
 
         return self.map(reduce_array, shortcut=shortcut)
 
 
-class DataArrayGroupBy(DataArrayGroupByBase, DataArrayGroupByReductions):
+# https://github.com/python/mypy/issues/9031
+class DataArrayGroupBy(  # type: ignore[misc]
+    DataArrayGroupByBase,
+    DataArrayGroupByReductions,
+    ImplementsArrayReduce,
+    IncludeCumMethods,
+):
     __slots__ = ()
 
 
-class DatasetGroupByBase(GroupBy, DatasetGroupbyArithmetic):
+class DatasetGroupByBase(GroupBy["Dataset"], DatasetGroupbyArithmetic):
 
     __slots__ = ()
+    _dims: Frozen[Hashable, int] | None
 
-    def map(self, func, args=(), shortcut=None, **kwargs):
+    @property
+    def dims(self) -> Frozen[Hashable, int]:
+        if self._dims is None:
+            self._dims = self._obj.isel({self._group_dim: self._group_indices[0]}).dims
+
+        return self._dims
+
+    def map(
+        self,
+        func: Callable[..., Dataset],
+        args: tuple[Any, ...] = (),
+        shortcut: bool | None = None,
+        **kwargs: Any,
+    ) -> Dataset:
         """Apply a function to each Dataset in the group and concatenate them
         together into a new Dataset.
 
@@ -942,7 +1260,7 @@ class DatasetGroupByBase(GroupBy, DatasetGroupbyArithmetic):
 
         Returns
         -------
-        applied : Dataset or DataArray
+        applied : Dataset
             The result of splitting, applying and combining this dataset.
         """
         # ignore shortcut if set (for now)
@@ -973,12 +1291,24 @@ class DatasetGroupByBase(GroupBy, DatasetGroupbyArithmetic):
         combined = _maybe_reorder(combined, dim, positions)
         # assign coord when the applied function does not return that coord
         if coord is not None and dim not in applied_example.dims:
-            combined[coord.name] = coord
+            index, index_vars = create_default_index_implicit(coord)
+            indexes = {k: index for k in index_vars}
+            combined = combined._overwrite_indexes(indexes, index_vars)
         combined = self._maybe_restore_empty_groups(combined)
         combined = self._maybe_unstack(combined)
         return combined
 
-    def reduce(self, func, dim=None, keep_attrs=None, **kwargs):
+    def reduce(
+        self,
+        func: Callable[..., Any],
+        dim: Dims | ellipsis = None,
+        *,
+        axis: int | Sequence[int] | None = None,
+        keep_attrs: bool | None = None,
+        keepdims: bool = False,
+        shortcut: bool = True,
+        **kwargs: Any,
+    ) -> Dataset:
         """Reduce the items in this group by applying `func` along some
         dimension(s).
 
@@ -988,8 +1318,9 @@ class DatasetGroupByBase(GroupBy, DatasetGroupbyArithmetic):
             Function which can be called in the form
             `func(x, axis=axis, **kwargs)` to return the result of collapsing
             an np.ndarray over an integer valued axis.
-        dim : ..., str or sequence of str, optional
-            Dimension(s) over which to apply `func`.
+        dim : ..., str, Iterable of Hashable or None, optional
+            Dimension(s) over which to apply `func`. By default apply over the
+            groupby dimension, with "..." apply over all dimensions.
         axis : int or sequence of int, optional
             Axis(es) over which to apply `func`. Only one of the 'dimension'
             and 'axis' arguments can be supplied. If neither are supplied, then
@@ -1003,24 +1334,31 @@ class DatasetGroupByBase(GroupBy, DatasetGroupbyArithmetic):
 
         Returns
         -------
-        reduced : Array
+        reduced : Dataset
             Array with summarized data and the indicated dimension(s)
             removed.
         """
         if dim is None:
-            dim = self._group_dim
+            dim = [self._group_dim]
 
         if keep_attrs is None:
-            keep_attrs = _get_keep_attrs(default=False)
+            keep_attrs = _get_keep_attrs(default=True)
 
-        def reduce_dataset(ds):
-            return ds.reduce(func, dim, keep_attrs, **kwargs)
+        def reduce_dataset(ds: Dataset) -> Dataset:
+            return ds.reduce(
+                func=func,
+                dim=dim,
+                axis=axis,
+                keep_attrs=keep_attrs,
+                keepdims=keepdims,
+                **kwargs,
+            )
 
         check_reduce_dims(dim, self.dims)
 
         return self.map(reduce_dataset)
 
-    def assign(self, **kwargs):
+    def assign(self, **kwargs: Any) -> Dataset:
         """Assign data variables by group.
 
         See Also
@@ -1030,5 +1368,11 @@ class DatasetGroupByBase(GroupBy, DatasetGroupbyArithmetic):
         return self.map(lambda ds: ds.assign(**kwargs))
 
 
-class DatasetGroupBy(DatasetGroupByBase, DatasetGroupByReductions):
+# https://github.com/python/mypy/issues/9031
+class DatasetGroupBy(  # type: ignore[misc]
+    DatasetGroupByBase,
+    DatasetGroupByReductions,
+    ImplementsDatasetReduce,
+    IncludeCumMethods,
+):
     __slots__ = ()
