@@ -1,7 +1,10 @@
+from __future__ import annotations
+
 import importlib
 import platform
 import warnings
 from contextlib import contextmanager, nullcontext
+from typing import Any
 from unittest import mock  # noqa: F401
 
 import numpy as np
@@ -40,7 +43,7 @@ arm_xfail = pytest.mark.xfail(
 )
 
 
-def _importorskip(modname, minversion=None):
+def _importorskip(modname: str, minversion: str | None = None) -> tuple[bool, Any]:
     try:
         mod = importlib.import_module(modname)
         has = True
@@ -62,7 +65,6 @@ has_h5netcdf_0_12, requires_h5netcdf_0_12 = _importorskip("h5netcdf", minversion
 has_pynio, requires_pynio = _importorskip("Nio")
 has_pseudonetcdf, requires_pseudonetcdf = _importorskip("PseudoNetCDF")
 has_cftime, requires_cftime = _importorskip("cftime")
-has_cftime_1_4_1, requires_cftime_1_4_1 = _importorskip("cftime", minversion="1.4.1")
 has_dask, requires_dask = _importorskip("dask")
 has_bottleneck, requires_bottleneck = _importorskip("bottleneck")
 has_nc_time_axis, requires_nc_time_axis = _importorskip("nc_time_axis")
@@ -78,6 +80,8 @@ has_cupy, requires_cupy = _importorskip("cupy")
 has_cartopy, requires_cartopy = _importorskip("cartopy")
 has_pint, requires_pint = _importorskip("pint")
 has_numexpr, requires_numexpr = _importorskip("numexpr")
+has_flox, requires_flox = _importorskip("flox")
+
 
 # some special cases
 has_scipy_or_netCDF4 = has_scipy or has_netCDF4
@@ -177,28 +181,28 @@ def assert_no_warnings():
 # invariants
 
 
-def assert_equal(a, b):
+def assert_equal(a, b, check_default_indexes=True):
     __tracebackhide__ = True
     xarray.testing.assert_equal(a, b)
-    xarray.testing._assert_internal_invariants(a)
-    xarray.testing._assert_internal_invariants(b)
+    xarray.testing._assert_internal_invariants(a, check_default_indexes)
+    xarray.testing._assert_internal_invariants(b, check_default_indexes)
 
 
-def assert_identical(a, b):
+def assert_identical(a, b, check_default_indexes=True):
     __tracebackhide__ = True
     xarray.testing.assert_identical(a, b)
-    xarray.testing._assert_internal_invariants(a)
-    xarray.testing._assert_internal_invariants(b)
+    xarray.testing._assert_internal_invariants(a, check_default_indexes)
+    xarray.testing._assert_internal_invariants(b, check_default_indexes)
 
 
-def assert_allclose(a, b, **kwargs):
+def assert_allclose(a, b, check_default_indexes=True, **kwargs):
     __tracebackhide__ = True
     xarray.testing.assert_allclose(a, b, **kwargs)
-    xarray.testing._assert_internal_invariants(a)
-    xarray.testing._assert_internal_invariants(b)
+    xarray.testing._assert_internal_invariants(a, check_default_indexes)
+    xarray.testing._assert_internal_invariants(b, check_default_indexes)
 
 
-def create_test_data(seed=None, add_attrs=True):
+def create_test_data(seed: int | None = None, add_attrs: bool = True) -> Dataset:
     rs = np.random.RandomState(seed)
     _vars = {
         "var1": ["dim1", "dim2"],
