@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import warnings
 from datetime import timedelta
 from itertools import product
@@ -34,9 +36,7 @@ from . import (
     assert_array_equal,
     assert_no_warnings,
     has_cftime,
-    has_cftime_1_4_1,
     requires_cftime,
-    requires_cftime_1_4_1,
     requires_dask,
 )
 
@@ -1007,12 +1007,9 @@ def test_decode_ambiguous_time_warns(calendar) -> None:
     units = "days since 1-1-1"
     expected = num2date(dates, units, calendar=calendar, only_use_cftime_datetimes=True)
 
-    exp_warn_type = SerializationWarning if is_standard_calendar else None
-
-    with pytest.warns(exp_warn_type) as record:
-        result = decode_cf_datetime(dates, units, calendar=calendar)
-
     if is_standard_calendar:
+        with pytest.warns(SerializationWarning) as record:
+            result = decode_cf_datetime(dates, units, calendar=calendar)
         relevant_warnings = [
             r
             for r in record.list
@@ -1020,7 +1017,8 @@ def test_decode_ambiguous_time_warns(calendar) -> None:
         ]
         assert len(relevant_warnings) == 1
     else:
-        assert not record
+        with assert_no_warnings():
+            result = decode_cf_datetime(dates, units, calendar=calendar)
 
     np.testing.assert_array_equal(result, expected)
 
@@ -1031,8 +1029,8 @@ def test_decode_ambiguous_time_warns(calendar) -> None:
 def test_encode_cf_datetime_defaults_to_correct_dtype(
     encoding_units, freq, date_range
 ) -> None:
-    if not has_cftime_1_4_1 and date_range == cftime_range:
-        pytest.skip("Test requires cftime 1.4.1.")
+    if not has_cftime and date_range == cftime_range:
+        pytest.skip("Test requires cftime")
     if (freq == "N" or encoding_units == "nanoseconds") and date_range == cftime_range:
         pytest.skip("Nanosecond frequency is not valid for cftime dates.")
     times = date_range("2000", periods=3, freq=freq)
@@ -1059,7 +1057,7 @@ def test_encode_decode_roundtrip_datetime64(freq) -> None:
     assert_equal(variable, decoded)
 
 
-@requires_cftime_1_4_1
+@requires_cftime
 @pytest.mark.parametrize("freq", ["U", "L", "S", "T", "H", "D"])
 def test_encode_decode_roundtrip_cftime(freq) -> None:
     initial_time = cftime_range("0001", periods=1)
@@ -1123,3 +1121,55 @@ def test_should_cftime_be_used_target_not_npable():
         ValueError, match="Calendar 'noleap' is only valid with cftime."
     ):
         _should_cftime_be_used(src, "noleap", False)
+
+
+@pytest.mark.parametrize("dtype", [np.uint8, np.uint16, np.uint32, np.uint64])
+def test_decode_cf_datetime_uint(dtype):
+    units = "seconds since 2018-08-22T03:23:03Z"
+    num_dates = dtype(50)
+    result = decode_cf_datetime(num_dates, units)
+    expected = np.asarray(np.datetime64("2018-08-22T03:23:53", "ns"))
+    np.testing.assert_equal(result, expected)
+
+
+@requires_cftime
+def test_decode_cf_datetime_uint64_with_cftime():
+    units = "days since 1700-01-01"
+    num_dates = np.uint64(182621)
+    result = decode_cf_datetime(num_dates, units)
+    expected = np.asarray(np.datetime64("2200-01-01", "ns"))
+    np.testing.assert_equal(result, expected)
+
+
+@requires_cftime
+def test_decode_cf_datetime_uint64_with_cftime_overflow_error():
+    units = "microseconds since 1700-01-01"
+    calendar = "360_day"
+    num_dates = np.uint64(1_000_000 * 86_400 * 360 * 500_000)
+    with pytest.raises(OverflowError):
+        decode_cf_datetime(num_dates, units, calendar)
+
+
+@pytest.mark.parametrize("use_cftime", [True, False])
+def test_decode_0size_datetime(use_cftime):
+    # GH1329
+    if use_cftime and not has_cftime:
+        pytest.skip()
+
+    dtype = object if use_cftime else "M8[ns]"
+    expected = np.array([], dtype=dtype)
+    actual = decode_cf_datetime(
+        np.zeros(shape=0, dtype=np.int64),
+        units="days since 1970-01-01 00:00:00",
+        calendar="proleptic_gregorian",
+        use_cftime=use_cftime,
+    )
+    np.testing.assert_equal(expected, actual)
+
+
+@requires_cftime
+def test_scalar_unit() -> None:
+    # test that a scalar units (often NaN when using to_netcdf) does not raise an error
+    variable = Variable(("x", "y"), np.array([[0, 1], [2, 3]]), {"units": np.nan})
+    result = coding.times.CFDatetimeCoder().decode(variable)
+    assert np.isnan(result.attrs["units"])
