@@ -1,4 +1,7 @@
+from __future__ import annotations
+
 from copy import deepcopy
+from typing import TYPE_CHECKING, Any
 
 import numpy as np
 import pandas as pd
@@ -6,6 +9,7 @@ import pytest
 
 from xarray import DataArray, Dataset, Variable, concat
 from xarray.core import dtypes, merge
+from xarray.core.indexes import PandasIndex
 
 from . import (
     InaccessibleArray,
@@ -16,8 +20,11 @@ from . import (
 )
 from .test_dataset import create_test_data
 
+if TYPE_CHECKING:
+    from xarray.core.types import CombineAttrsOptions, JoinOptions
 
-def test_concat_compat():
+
+def test_concat_compat() -> None:
     ds1 = Dataset(
         {
             "has_x_y": (("y", "x"), [[1, 2]]),
@@ -50,10 +57,10 @@ def test_concat_compat():
 
 class TestConcatDataset:
     @pytest.fixture
-    def data(self):
+    def data(self) -> Dataset:
         return create_test_data().drop_dims("dim3")
 
-    def rectify_dim_order(self, data, dataset):
+    def rectify_dim_order(self, data, dataset) -> Dataset:
         # return a new dataset with all variable dimensions transposed into
         # the order in which they are found in `data`
         return Dataset(
@@ -64,11 +71,11 @@ class TestConcatDataset:
 
     @pytest.mark.parametrize("coords", ["different", "minimal"])
     @pytest.mark.parametrize("dim", ["dim1", "dim2"])
-    def test_concat_simple(self, data, dim, coords):
+    def test_concat_simple(self, data, dim, coords) -> None:
         datasets = [g for _, g in data.groupby(dim, squeeze=False)]
         assert_identical(data, concat(datasets, dim, coords=coords))
 
-    def test_concat_merge_variables_present_in_some_datasets(self, data):
+    def test_concat_merge_variables_present_in_some_datasets(self, data) -> None:
         # coordinates present in some datasets but not others
         ds1 = Dataset(data_vars={"a": ("y", [0.1])}, coords={"x": 0.1})
         ds2 = Dataset(data_vars={"a": ("y", [0.2])}, coords={"z": 0.2})
@@ -84,7 +91,7 @@ class TestConcatDataset:
         expected = data.copy().assign(foo=data1.foo)
         assert_identical(expected, actual)
 
-    def test_concat_2(self, data):
+    def test_concat_2(self, data) -> None:
         dim = "dim2"
         datasets = [g for _, g in data.groupby(dim, squeeze=True)]
         concat_over = [k for k, v in data.coords.items() if dim in v.dims and k != dim]
@@ -93,7 +100,7 @@ class TestConcatDataset:
 
     @pytest.mark.parametrize("coords", ["different", "minimal", "all"])
     @pytest.mark.parametrize("dim", ["dim1", "dim2"])
-    def test_concat_coords_kwarg(self, data, dim, coords):
+    def test_concat_coords_kwarg(self, data, dim, coords) -> None:
         data = data.copy(deep=True)
         # make sure the coords argument behaves as expected
         data.coords["extra"] = ("dim4", np.arange(3))
@@ -107,7 +114,7 @@ class TestConcatDataset:
         else:
             assert_equal(data["extra"], actual["extra"])
 
-    def test_concat(self, data):
+    def test_concat(self, data) -> None:
         split_data = [
             data.isel(dim1=slice(3)),
             data.isel(dim1=3),
@@ -115,7 +122,7 @@ class TestConcatDataset:
         ]
         assert_identical(data, concat(split_data, "dim1"))
 
-    def test_concat_dim_precedence(self, data):
+    def test_concat_dim_precedence(self, data) -> None:
         # verify that the dim argument takes precedence over
         # concatenating dataset variables of the same name
         dim = (2 * data["dim1"]).rename("dim1")
@@ -124,14 +131,23 @@ class TestConcatDataset:
         expected["dim1"] = dim
         assert_identical(expected, concat(datasets, dim))
 
-    def test_concat_data_vars(self):
+    def test_concat_data_vars_typing(self) -> None:
+        # Testing typing, can be removed if the next function works with annotations.
         data = Dataset({"foo": ("x", np.random.randn(10))})
-        objs = [data.isel(x=slice(5)), data.isel(x=slice(5, None))]
+        objs: list[Dataset] = [data.isel(x=slice(5)), data.isel(x=slice(5, None))]
+        actual = concat(objs, dim="x", data_vars="minimal")
+        assert_identical(data, actual)
+
+    def test_concat_data_vars(self):
+        # TODO: annotating this func fails
+        data = Dataset({"foo": ("x", np.random.randn(10))})
+        objs: list[Dataset] = [data.isel(x=slice(5)), data.isel(x=slice(5, None))]
         for data_vars in ["minimal", "different", "all", [], ["foo"]]:
             actual = concat(objs, dim="x", data_vars=data_vars)
             assert_identical(data, actual)
 
     def test_concat_coords(self):
+        # TODO: annotating this func fails
         data = Dataset({"foo": ("x", np.random.randn(10))})
         expected = data.assign_coords(c=("x", [0] * 5 + [1] * 5))
         objs = [
@@ -146,6 +162,7 @@ class TestConcatDataset:
                 concat(objs, dim="x", coords=coords)
 
     def test_concat_constant_index(self):
+        # TODO: annotating this func fails
         # GH425
         ds1 = Dataset({"foo": 1.5}, {"y": 1})
         ds2 = Dataset({"foo": 2.5}, {"y": 1})
@@ -158,7 +175,7 @@ class TestConcatDataset:
             # "foo" has dimension "y" so minimal should concatenate it?
             concat([ds1, ds2], "new_dim", data_vars="minimal")
 
-    def test_concat_size0(self):
+    def test_concat_size0(self) -> None:
         data = create_test_data()
         split_data = [data.isel(dim1=slice(0, 0)), data]
         actual = concat(split_data, "dim1")
@@ -167,7 +184,7 @@ class TestConcatDataset:
         actual = concat(split_data[::-1], "dim1")
         assert_identical(data, actual)
 
-    def test_concat_autoalign(self):
+    def test_concat_autoalign(self) -> None:
         ds1 = Dataset({"foo": DataArray([1, 2], coords=[("x", [1, 2])])})
         ds2 = Dataset({"foo": DataArray([1, 2], coords=[("x", [1, 3])])})
         actual = concat([ds1, ds2], "y")
@@ -183,6 +200,7 @@ class TestConcatDataset:
         assert_identical(expected, actual)
 
     def test_concat_errors(self):
+        # TODO: annotating this func fails
         data = create_test_data()
         split_data = [data.isel(dim1=slice(3)), data.isel(dim1=slice(3, None))]
 
@@ -201,7 +219,9 @@ class TestConcatDataset:
             concat([data, data], "new_dim", coords=["not_found"])
 
         with pytest.raises(ValueError, match=r"global attributes not"):
-            data0, data1 = deepcopy(split_data)
+            # call deepcopy seperately to get unique attrs
+            data0 = deepcopy(split_data[0])
+            data1 = deepcopy(split_data[1])
             data1.attrs["foo"] = "bar"
             concat([data0, data1], "dim1", compat="identical")
         assert_identical(data, concat([data0, data1], "dim1", compat="equals"))
@@ -222,11 +242,11 @@ class TestConcatDataset:
         ):
             concat([Dataset({"x": 0}), Dataset({}, {"x": 1})], dim="z")
 
-    def test_concat_join_kwarg(self):
+    def test_concat_join_kwarg(self) -> None:
         ds1 = Dataset({"a": (("x", "y"), [[0]])}, coords={"x": [0], "y": [0]})
         ds2 = Dataset({"a": (("x", "y"), [[0]])}, coords={"x": [1], "y": [0.0001]})
 
-        expected = {}
+        expected: dict[JoinOptions, Any] = {}
         expected["outer"] = Dataset(
             {"a": (("x", "y"), [[0, np.nan], [np.nan, 0]])},
             {"x": [0, 1], "y": [0, 0.0001]},
@@ -247,7 +267,7 @@ class TestConcatDataset:
             coords={"x": [0, 1], "y": [0]},
         )
 
-        with pytest.raises(ValueError, match=r"indexes along dimension 'y'"):
+        with pytest.raises(ValueError, match=r"cannot align.*exact.*dimensions.*'y'"):
             actual = concat([ds1, ds2], join="exact", dim="x")
 
         for join in expected:
@@ -258,10 +278,10 @@ class TestConcatDataset:
         actual = concat(
             [ds1.drop_vars("x"), ds2.drop_vars("x")], join="override", dim="y"
         )
-        expected = Dataset(
+        expected2 = Dataset(
             {"a": (("x", "y"), np.array([0, 0], ndmin=2))}, coords={"y": [0, 0.0001]}
         )
-        assert_identical(actual, expected)
+        assert_identical(actual, expected2)
 
     @pytest.mark.parametrize(
         "combine_attrs, var1_attrs, var2_attrs, expected_attrs, expect_exception",
@@ -299,6 +319,13 @@ class TestConcatDataset:
                 {"a": 41, "c": 43, "d": 44},
                 False,
             ),
+            (
+                lambda attrs, context: {"a": -1, "b": 0, "c": 1} if any(attrs) else {},
+                {"a": 41, "b": 42, "c": 43},
+                {"b": 2, "c": 43, "d": 44},
+                {"a": -1, "b": 0, "c": 1},
+                False,
+            ),
         ],
     )
     def test_concat_combine_attrs_kwarg(
@@ -354,6 +381,13 @@ class TestConcatDataset:
                 {"a": 41, "c": 43, "d": 44},
                 False,
             ),
+            (
+                lambda attrs, context: {"a": -1, "b": 0, "c": 1} if any(attrs) else {},
+                {"a": 41, "b": 42, "c": 43},
+                {"b": 2, "c": 43, "d": 44},
+                {"a": -1, "b": 0, "c": 1},
+                False,
+            ),
         ],
     )
     def test_concat_combine_attrs_kwarg_variables(
@@ -375,7 +409,7 @@ class TestConcatDataset:
 
             assert_identical(actual, expected)
 
-    def test_concat_promote_shape(self):
+    def test_concat_promote_shape(self) -> None:
         # mixed dims within variables
         objs = [Dataset({}, {"x": 0}), Dataset({"x": [1]})]
         actual = concat(objs, "x")
@@ -413,7 +447,35 @@ class TestConcatDataset:
         expected = Dataset({"z": (("x", "y"), [[-1], [1]])}, {"x": [0, 1], "y": [0]})
         assert_identical(actual, expected)
 
-    def test_concat_do_not_promote(self):
+        # regression GH6384
+        objs = [
+            Dataset({}, {"x": pd.Interval(-1, 0, closed="right")}),
+            Dataset({"x": [pd.Interval(0, 1, closed="right")]}),
+        ]
+        actual = concat(objs, "x")
+        expected = Dataset(
+            {
+                "x": [
+                    pd.Interval(-1, 0, closed="right"),
+                    pd.Interval(0, 1, closed="right"),
+                ]
+            }
+        )
+        assert_identical(actual, expected)
+
+        # regression GH6416 (coord dtype) and GH6434
+        time_data1 = np.array(["2022-01-01", "2022-02-01"], dtype="datetime64[ns]")
+        time_data2 = np.array("2022-03-01", dtype="datetime64[ns]")
+        time_expected = np.array(
+            ["2022-01-01", "2022-02-01", "2022-03-01"], dtype="datetime64[ns]"
+        )
+        objs = [Dataset({}, {"time": time_data1}), Dataset({}, {"time": time_data2})]
+        actual = concat(objs, "time")
+        expected = Dataset({}, {"time": time_expected})
+        assert_identical(actual, expected)
+        assert isinstance(actual.indexes["time"], pd.DatetimeIndex)
+
+    def test_concat_do_not_promote(self) -> None:
         # GH438
         objs = [
             Dataset({"y": ("t", [1])}, {"x": 1, "t": [0]}),
@@ -430,24 +492,41 @@ class TestConcatDataset:
         with pytest.raises(ValueError):
             concat(objs, "t", coords="minimal")
 
-    def test_concat_dim_is_variable(self):
+    def test_concat_dim_is_variable(self) -> None:
         objs = [Dataset({"x": 0}), Dataset({"x": 1})]
-        coord = Variable("y", [3, 4])
-        expected = Dataset({"x": ("y", [0, 1]), "y": [3, 4]})
+        coord = Variable("y", [3, 4], attrs={"foo": "bar"})
+        expected = Dataset({"x": ("y", [0, 1]), "y": coord})
         actual = concat(objs, coord)
         assert_identical(actual, expected)
 
-    def test_concat_multiindex(self):
+    def test_concat_dim_is_dataarray(self) -> None:
+        objs = [Dataset({"x": 0}), Dataset({"x": 1})]
+        coord = DataArray([3, 4], dims="y", attrs={"foo": "bar"})
+        expected = Dataset({"x": ("y", [0, 1]), "y": coord})
+        actual = concat(objs, coord)
+        assert_identical(actual, expected)
+
+    def test_concat_multiindex(self) -> None:
         x = pd.MultiIndex.from_product([[1, 2, 3], ["a", "b"]])
-        expected = Dataset({"x": x})
+        expected = Dataset(coords={"x": x})
         actual = concat(
             [expected.isel(x=slice(2)), expected.isel(x=slice(2, None))], "x"
         )
         assert expected.equals(actual)
         assert isinstance(actual.x.to_index(), pd.MultiIndex)
 
+    def test_concat_along_new_dim_multiindex(self) -> None:
+        # see https://github.com/pydata/xarray/issues/6881
+        level_names = ["x_level_0", "x_level_1"]
+        x = pd.MultiIndex.from_product([[1, 2, 3], ["a", "b"]], names=level_names)
+        ds = Dataset(coords={"x": x})
+        concatenated = concat([ds], "new")
+        actual = list(concatenated.xindexes.get_all_coords("x"))
+        expected = ["x"] + level_names
+        assert actual == expected
+
     @pytest.mark.parametrize("fill_value", [dtypes.NA, 2, 2.0, {"a": 2, "b": 1}])
-    def test_concat_fill_value(self, fill_value):
+    def test_concat_fill_value(self, fill_value) -> None:
         datasets = [
             Dataset({"a": ("x", [2, 3]), "b": ("x", [-2, 1]), "x": [1, 2]}),
             Dataset({"a": ("x", [1, 2]), "b": ("x", [3, -1]), "x": [0, 1]}),
@@ -473,7 +552,7 @@ class TestConcatDataset:
 
     @pytest.mark.parametrize("dtype", [str, bytes])
     @pytest.mark.parametrize("dim", ["x1", "x2"])
-    def test_concat_str_dtype(self, dtype, dim):
+    def test_concat_str_dtype(self, dtype, dim) -> None:
 
         data = np.arange(4).reshape([2, 2])
 
@@ -497,7 +576,7 @@ class TestConcatDataset:
 
 
 class TestConcatDataArray:
-    def test_concat(self):
+    def test_concat(self) -> None:
         ds = Dataset(
             {
                 "foo": (["x", "y"], np.random.random((2, 3))),
@@ -521,16 +600,16 @@ class TestConcatDataArray:
         stacked = concat(grouped, ds["x"])
         assert_identical(foo, stacked)
         # with an index as the 'dim' argument
-        stacked = concat(grouped, ds.indexes["x"])
+        stacked = concat(grouped, pd.Index(ds["x"], name="x"))
         assert_identical(foo, stacked)
 
-        actual = concat([foo[0], foo[1]], pd.Index([0, 1])).reset_coords(drop=True)
+        actual2 = concat([foo[0], foo[1]], pd.Index([0, 1])).reset_coords(drop=True)
         expected = foo[:2].rename({"x": "concat_dim"})
-        assert_identical(expected, actual)
+        assert_identical(expected, actual2)
 
-        actual = concat([foo[0], foo[1]], [0, 1]).reset_coords(drop=True)
+        actual3 = concat([foo[0], foo[1]], [0, 1]).reset_coords(drop=True)
         expected = foo[:2].rename({"x": "concat_dim"})
-        assert_identical(expected, actual)
+        assert_identical(expected, actual3)
 
         with pytest.raises(ValueError, match=r"not identical"):
             concat([foo, bar], dim="w", compat="identical")
@@ -538,7 +617,7 @@ class TestConcatDataArray:
         with pytest.raises(ValueError, match=r"not a valid argument"):
             concat([foo, bar], dim="w", data_vars="minimal")
 
-    def test_concat_encoding(self):
+    def test_concat_encoding(self) -> None:
         # Regression test for GH1297
         ds = Dataset(
             {
@@ -554,7 +633,7 @@ class TestConcatDataArray:
         assert concat([ds, ds], dim="x").encoding == ds.encoding
 
     @requires_dask
-    def test_concat_lazy(self):
+    def test_concat_lazy(self) -> None:
         import dask.array as da
 
         arrays = [
@@ -569,7 +648,7 @@ class TestConcatDataArray:
         assert combined.dims == ("z", "x", "y")
 
     @pytest.mark.parametrize("fill_value", [dtypes.NA, 2, 2.0])
-    def test_concat_fill_value(self, fill_value):
+    def test_concat_fill_value(self, fill_value) -> None:
         foo = DataArray([1, 2], coords=[("x", [1, 2])])
         bar = DataArray([1, 2], coords=[("x", [1, 3])])
         if fill_value == dtypes.NA:
@@ -584,7 +663,7 @@ class TestConcatDataArray:
         actual = concat((foo, bar), dim="y", fill_value=fill_value)
         assert_identical(actual, expected)
 
-    def test_concat_join_kwarg(self):
+    def test_concat_join_kwarg(self) -> None:
         ds1 = Dataset(
             {"a": (("x", "y"), [[0]])}, coords={"x": [0], "y": [0]}
         ).to_array()
@@ -592,7 +671,7 @@ class TestConcatDataArray:
             {"a": (("x", "y"), [[0]])}, coords={"x": [1], "y": [0.0001]}
         ).to_array()
 
-        expected = {}
+        expected: dict[JoinOptions, Any] = {}
         expected["outer"] = Dataset(
             {"a": (("x", "y"), [[0, np.nan], [np.nan, 0]])},
             {"x": [0, 1], "y": [0, 0.0001]},
@@ -613,18 +692,18 @@ class TestConcatDataArray:
             coords={"x": [0, 1], "y": [0]},
         )
 
-        with pytest.raises(ValueError, match=r"indexes along dimension 'y'"):
+        with pytest.raises(ValueError, match=r"cannot align.*exact.*dimensions.*'y'"):
             actual = concat([ds1, ds2], join="exact", dim="x")
 
         for join in expected:
             actual = concat([ds1, ds2], join=join, dim="x")
             assert_equal(actual, expected[join].to_array())
 
-    def test_concat_combine_attrs_kwarg(self):
+    def test_concat_combine_attrs_kwarg(self) -> None:
         da1 = DataArray([0], coords=[("x", [0])], attrs={"b": 42})
         da2 = DataArray([0], coords=[("x", [1])], attrs={"b": 42, "c": 43})
 
-        expected = {}
+        expected: dict[CombineAttrsOptions, Any] = {}
         expected["drop"] = DataArray([0, 0], coords=[("x", [0, 1])])
         expected["no_conflicts"] = DataArray(
             [0, 0], coords=[("x", [0, 1])], attrs={"b": 42, "c": 43}
@@ -646,7 +725,7 @@ class TestConcatDataArray:
 
     @pytest.mark.parametrize("dtype", [str, bytes])
     @pytest.mark.parametrize("dim", ["x1", "x2"])
-    def test_concat_str_dtype(self, dtype, dim):
+    def test_concat_str_dtype(self, dtype, dim) -> None:
 
         data = np.arange(4).reshape([2, 2])
 
@@ -664,10 +743,19 @@ class TestConcatDataArray:
 
         assert np.issubdtype(actual.x2.dtype, dtype)
 
+    def test_concat_coord_name(self) -> None:
+
+        da = DataArray([0], dims="a")
+        da_concat = concat([da, da], dim=DataArray([0, 1], dims="b"))
+        assert list(da_concat.coords) == ["b"]
+
+        da_concat_std = concat([da, da], dim=DataArray([0, 1]))
+        assert list(da_concat_std.coords) == ["dim_0"]
+
 
 @pytest.mark.parametrize("attr1", ({"a": {"meta": [10, 20, 30]}}, {"a": [1, 2, 3]}, {}))
 @pytest.mark.parametrize("attr2", ({"a": [1, 2, 3]}, {}))
-def test_concat_attrs_first_variable(attr1, attr2):
+def test_concat_attrs_first_variable(attr1, attr2) -> None:
 
     arrs = [
         DataArray([[1], [2]], dims=["x", "y"], attrs=attr1),
@@ -679,6 +767,7 @@ def test_concat_attrs_first_variable(attr1, attr2):
 
 
 def test_concat_merge_single_non_dim_coord():
+    # TODO: annotating this func fails
     da1 = DataArray([1, 2, 3], dims="x", coords={"x": [1, 2, 3], "y": 1})
     da2 = DataArray([4, 5, 6], dims="x", coords={"x": [4, 5, 6]})
 
@@ -699,7 +788,7 @@ def test_concat_merge_single_non_dim_coord():
             concat([da1, da2, da3], dim="x")
 
 
-def test_concat_preserve_coordinate_order():
+def test_concat_preserve_coordinate_order() -> None:
     x = np.arange(0, 5)
     y = np.arange(0, 10)
     time = np.arange(0, 4)
@@ -730,3 +819,44 @@ def test_concat_preserve_coordinate_order():
     for act, exp in zip(actual.coords, expected.coords):
         assert act == exp
         assert_identical(actual.coords[act], expected.coords[exp])
+
+
+def test_concat_typing_check() -> None:
+    ds = Dataset({"foo": 1}, {"bar": 2})
+    da = Dataset({"foo": 3}, {"bar": 4}).to_array(dim="foo")
+
+    # concatenate a list of non-homogeneous types must raise TypeError
+    with pytest.raises(
+        TypeError,
+        match="The elements in the input list need to be either all 'Dataset's or all 'DataArray's",
+    ):
+        concat([ds, da], dim="foo")  # type: ignore
+    with pytest.raises(
+        TypeError,
+        match="The elements in the input list need to be either all 'Dataset's or all 'DataArray's",
+    ):
+        concat([da, ds], dim="foo")  # type: ignore
+
+
+def test_concat_not_all_indexes() -> None:
+    ds1 = Dataset(coords={"x": ("x", [1, 2])})
+    # ds2.x has no default index
+    ds2 = Dataset(coords={"x": ("y", [3, 4])})
+
+    with pytest.raises(
+        ValueError, match=r"'x' must have either an index or no index in all datasets.*"
+    ):
+        concat([ds1, ds2], dim="x")
+
+
+def test_concat_index_not_same_dim() -> None:
+    ds1 = Dataset(coords={"x": ("x", [1, 2])})
+    ds2 = Dataset(coords={"x": ("y", [3, 4])})
+    # TODO: use public API for setting a non-default index, when available
+    ds2._indexes["x"] = PandasIndex([3, 4], "y")
+
+    with pytest.raises(
+        ValueError,
+        match=r"Cannot concatenate along dimension 'x' indexes with dimensions.*",
+    ):
+        concat([ds1, ds2], dim="x")
