@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import sys
 from textwrap import dedent
 
@@ -9,7 +11,7 @@ from numpy.core import defchararray
 import xarray as xr
 from xarray.core import formatting
 
-from . import requires_netCDF4
+from . import requires_dask, requires_netCDF4
 
 
 class TestFormatting:
@@ -188,6 +190,11 @@ class TestFormatting:
     def test_maybe_truncate(self) -> None:
         assert formatting.maybe_truncate("ß", 10) == "ß"
 
+    def test_format_timestamp_invalid_pandas_format(self) -> None:
+        expected = "2021-12-06 17:00:00 00"
+        with pytest.raises(ValueError):
+            formatting.format_timestamp(expected)
+
     def test_format_timestamp_out_of_bounds(self) -> None:
         from datetime import datetime
 
@@ -384,7 +391,10 @@ class TestFormatting:
     def test_array_repr(self) -> None:
         ds = xr.Dataset(coords={"foo": [1, 2, 3], "bar": [1, 2, 3]})
         ds[(1, 2)] = xr.DataArray([0], dims="test")
-        actual = formatting.array_repr(ds[(1, 2)])
+        ds_12 = ds[(1, 2)]
+
+        # Test repr function behaves correctly:
+        actual = formatting.array_repr(ds_12)
         expected = dedent(
             """\
         <xarray.DataArray (1, 2) (test: 1)>
@@ -394,6 +404,14 @@ class TestFormatting:
 
         assert actual == expected
 
+        # Test repr, str prints returns correctly as well:
+        assert repr(ds_12) == expected
+        assert str(ds_12) == expected
+
+        # f-strings (aka format(...)) by default should use the repr:
+        actual = f"{ds_12}"
+        assert actual == expected
+
         with xr.set_options(display_expand_data=False):
             actual = formatting.array_repr(ds[(1, 2)])
             expected = dedent(
@@ -413,6 +431,47 @@ class TestFormatting:
         with xr.set_options(display_expand_data=False):
             formatting.array_repr(var)
 
+    def test_array_repr_recursive(self) -> None:
+        # GH:issue:7111
+
+        # direct recurion
+        var = xr.Variable("x", [0, 1])
+        var.attrs["x"] = var
+        formatting.array_repr(var)
+
+        da = xr.DataArray([0, 1], dims=["x"])
+        da.attrs["x"] = da
+        formatting.array_repr(da)
+
+        # indirect recursion
+        var.attrs["x"] = da
+        da.attrs["x"] = var
+        formatting.array_repr(var)
+        formatting.array_repr(da)
+
+    @requires_dask
+    def test_array_scalar_format(self) -> None:
+        # Test numpy scalars:
+        var = xr.DataArray(np.array(0))
+        assert format(var, "") == repr(var)
+        assert format(var, "d") == "0"
+        assert format(var, ".2f") == "0.00"
+
+        # Test dask scalars, not supported however:
+        import dask.array as da
+
+        var = xr.DataArray(da.array(0))
+        assert format(var, "") == repr(var)
+        with pytest.raises(TypeError) as excinfo:
+            format(var, ".2f")
+        assert "unsupported format string passed to" in str(excinfo.value)
+
+        # Test numpy arrays raises:
+        var = xr.DataArray([0.1, 0.2])
+        with pytest.raises(NotImplementedError) as excinfo:  # type: ignore
+            format(var, ".2f")
+        assert "Using format_spec is only supported" in str(excinfo.value)
+
 
 def test_inline_variable_array_repr_custom_repr() -> None:
     class CustomArray:
@@ -431,7 +490,7 @@ def test_inline_variable_array_repr_custom_repr() -> None:
             return NotImplemented
 
         @property
-        def shape(self):
+        def shape(self) -> tuple[int, ...]:
             return self.value.shape
 
         @property
@@ -474,6 +533,12 @@ def test_short_numpy_repr() -> None:
         num_lines = formatting.short_numpy_repr(array).count("\n") + 1
         assert num_lines < 30
 
+    # threshold option (default: 200)
+    array2 = np.arange(100)
+    assert "..." not in formatting.short_numpy_repr(array2)
+    with xr.set_options(display_values_threshold=10):
+        assert "..." in formatting.short_numpy_repr(array2)
+
 
 def test_large_array_repr_length() -> None:
 
@@ -547,18 +612,65 @@ def test__mapping_repr(display_max_rows, n_vars, n_attr) -> None:
         assert len_summary == n_vars
 
     with xr.set_options(
+        display_max_rows=display_max_rows,
         display_expand_coords=False,
         display_expand_data_vars=False,
         display_expand_attrs=False,
     ):
         actual = formatting.dataset_repr(ds)
-        coord_s = ", ".join([f"{c}: {len(v)}" for c, v in coords.items()])
-        expected = dedent(
-            f"""\
-            <xarray.Dataset>
-            Dimensions:      ({coord_s})
-            Coordinates: ({n_vars})
-            Data variables: ({n_vars})
-            Attributes: ({n_attr})"""
+        col_width = formatting._calculate_col_width(ds.variables)
+        dims_start = formatting.pretty_print("Dimensions:", col_width)
+        dims_values = formatting.dim_summary_limited(
+            ds, col_width=col_width + 1, max_rows=display_max_rows
         )
+        expected = f"""\
+<xarray.Dataset>
+{dims_start}({dims_values})
+Coordinates: ({n_vars})
+Data variables: ({n_vars})
+Attributes: ({n_attr})"""
+        expected = dedent(expected)
         assert actual == expected
+
+
+def test__mapping_repr_recursive() -> None:
+    # GH:issue:7111
+
+    # direct recursion
+    ds = xr.Dataset({"a": [["x"], [1, 2, 3]]})
+    ds.attrs["ds"] = ds
+    formatting.dataset_repr(ds)
+
+    # indirect recursion
+    ds2 = xr.Dataset({"b": [["y"], [1, 2, 3]]})
+    ds.attrs["ds"] = ds2
+    ds2.attrs["ds"] = ds
+    formatting.dataset_repr(ds2)
+
+
+def test__element_formatter(n_elements: int = 100) -> None:
+    expected = """\
+    Dimensions without coordinates: dim_0: 3, dim_1: 3, dim_2: 3, dim_3: 3,
+                                    dim_4: 3, dim_5: 3, dim_6: 3, dim_7: 3,
+                                    dim_8: 3, dim_9: 3, dim_10: 3, dim_11: 3,
+                                    dim_12: 3, dim_13: 3, dim_14: 3, dim_15: 3,
+                                    dim_16: 3, dim_17: 3, dim_18: 3, dim_19: 3,
+                                    dim_20: 3, dim_21: 3, dim_22: 3, dim_23: 3,
+                                    ...
+                                    dim_76: 3, dim_77: 3, dim_78: 3, dim_79: 3,
+                                    dim_80: 3, dim_81: 3, dim_82: 3, dim_83: 3,
+                                    dim_84: 3, dim_85: 3, dim_86: 3, dim_87: 3,
+                                    dim_88: 3, dim_89: 3, dim_90: 3, dim_91: 3,
+                                    dim_92: 3, dim_93: 3, dim_94: 3, dim_95: 3,
+                                    dim_96: 3, dim_97: 3, dim_98: 3, dim_99: 3"""
+    expected = dedent(expected)
+
+    intro = "Dimensions without coordinates: "
+    elements = [
+        f"{k}: {v}" for k, v in {f"dim_{k}": 3 for k in np.arange(n_elements)}.items()
+    ]
+    values = xr.core.formatting._element_formatter(
+        elements, col_width=len(intro), max_rows=12
+    )
+    actual = intro + values
+    assert expected == actual
