@@ -1,19 +1,13 @@
-import sys
+from __future__ import annotations
+
+from importlib.metadata import EntryPoint
 from unittest import mock
 
 import pytest
 
 from xarray.backends import common, plugins
 
-if sys.version_info >= (3, 8):
-    from importlib.metadata import EntryPoint
-
-    importlib_metadata_mock = "importlib.metadata"
-else:
-    # if the fallback library is missing, we are doomed.
-    from importlib_metadata import EntryPoint
-
-    importlib_metadata_mock = "importlib_metadata"
+importlib_metadata_mock = "importlib.metadata"
 
 
 class DummyBackendEntrypointArgs(common.BackendEntrypoint):
@@ -91,7 +85,7 @@ def test_backends_dict_from_pkg() -> None:
     entrypoints = [EntryPoint(name, value, group) for name, value, group in specs]
     engines = plugins.backends_dict_from_pkg(entrypoints)
     assert len(engines) == 2
-    assert engines.keys() == set(("engine1", "engine2"))
+    assert engines.keys() == {"engine1", "engine2"}
 
 
 def test_set_missing_parameters() -> None:
@@ -153,8 +147,7 @@ def test_build_engines_sorted() -> None:
         EntryPoint("dummy2", "xarray.tests.test_plugins:backend_1", "xarray.backends"),
         EntryPoint("dummy1", "xarray.tests.test_plugins:backend_1", "xarray.backends"),
     ]
-    backend_entrypoints = plugins.build_engines(dummy_pkg_entrypoints)
-    backend_entrypoints = list(backend_entrypoints)
+    backend_entrypoints = list(plugins.build_engines(dummy_pkg_entrypoints))
 
     indices = []
     for be in plugins.STANDARD_BACKENDS_ORDER:
