@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import pytest
 
 from xarray import DataArray, tutorial
@@ -11,13 +13,15 @@ class TestLoadDataset:
     def setUp(self):
         self.testfile = "tiny"
 
-    def test_download_from_github(self, tmp_path):
+    def test_download_from_github(self, tmp_path) -> None:
         cache_dir = tmp_path / tutorial._default_cache_dir_name
         ds = tutorial.open_dataset(self.testfile, cache_dir=cache_dir).load()
         tiny = DataArray(range(5), name="tiny").to_dataset()
         assert_identical(ds, tiny)
 
-    def test_download_from_github_load_without_cache(self, tmp_path, monkeypatch):
+    def test_download_from_github_load_without_cache(
+        self, tmp_path, monkeypatch
+    ) -> None:
         cache_dir = tmp_path / tutorial._default_cache_dir_name
 
         ds_nocache = tutorial.open_dataset(
@@ -30,11 +34,11 @@ class TestLoadDataset:
         self, tmp_path, monkeypatch
     ):
         cache_dir = tmp_path / tutorial._default_cache_dir_name
-
-        arr_nocache = tutorial.open_rasterio(
-            "RGB.byte", cache=False, cache_dir=cache_dir
-        ).load()
-        arr_cache = tutorial.open_rasterio(
-            "RGB.byte", cache=True, cache_dir=cache_dir
-        ).load()
+        with pytest.warns(DeprecationWarning):
+            arr_nocache = tutorial.open_rasterio(
+                "RGB.byte", cache=False, cache_dir=cache_dir
+            ).load()
+            arr_cache = tutorial.open_rasterio(
+                "RGB.byte", cache=True, cache_dir=cache_dir
+            ).load()
         assert_identical(arr_cache, arr_nocache)
